import base64
import unicodedata
try:
    from hashlib import sha1
except ImportError:
    # Python < 2.5
    from sha import new as sha1

from twisted.words.protocols.jabber.client import IQ
from twisted.words.xish import domish, xpath

def make_iq_error(stream, stanza, cond=None):
    reply = IQ(stream, "error")
    reply["id"] = stanza["id"]
    reply["to"] = stanza["from"]
    reply["from"] = stanza["to"]
    error = reply.addElement("error")
    return reply

def make_iq_result(stream, stanza):
    reply = IQ(stream, "result")
    reply["id"] = stanza["id"]
    reply["to"] = stanza["from"]
    reply["from"] = stanza["to"]
    child = stanza.firstChildElement()

    if child:
        reply.addElement((child.uri, child.name))

    return reply

def xpath_query(query, elem):
    nodes = xpath.queryForNodes(query, elem)

    if nodes is None:
        return []
    else:
        return nodes

def _elem_add(elem, *children):
    for child in children:
        if isinstance(child, domish.Element):
            elem.addChild(child)
            if child.uri is None:
                # copy the parent's ns as addChild doesn't do it
                child.uri = elem.uri
        elif isinstance(child, unicode):
            elem.addContent(child)
        else:
            raise ValueError('invalid child object %r', child)

def elem(a, b=None, **kw):
    class _elem(domish.Element):
        def __call__(self, *children):
            _elem_add(self, *children)
            return self

    if b is not None:
        elem = _elem((a, b))
    else:
        elem = _elem((None, a))

    for k, v in kw.iteritems():
        if k == 'from_':
            elem['from'] = v
        else:
            elem[k] = v

    return elem

# FIXME: add test

def elem_iq(server, type, **kw):
    class _iq(IQ):
        def __call__(self, *children):
            _elem_add(self, *children)
            return self

    iq = _iq(server, type)

    for k, v in kw.iteritems():
        if k == 'from_':
            iq['from'] = v
        else:
            iq[k] = v

    return iq

def xml_pformat(e, indent=0, width=80):
    r"""
    >>> e = domish.Element((None, 'foo'))
    >>> print xml_pformat(e),
    <foo />
    >>> e['one'] = 'two'
    >>> print xml_pformat(e),
    <foo one="two" />
    >>> _ = e.addChild('hello')
    >>> print xml_pformat(e),
    <foo one="two">
      hello
    </foo>
    >>> bar = e.addElement((None, 'bar'))
    >>> print xml_pformat(e),
    <foo one="two">
      hello
      <bar />
    </foo>
    """

    prefix = ' ' * (indent + len(e.name) + 2)
    attributes = e.attributes.items()

    if e.uri:
        attributes.insert(0, ('xmlns', e.uri))

    ret = '%s<%s' % (' ' * indent, e.name)

    if attributes:
        ret += ' %s="%s"' % attributes[0]

        for name, value in attributes[1:]:
            ret += '\n%s%s="%s"' % (prefix, name, value)

    if e.children:
        ret += '>\n'

        for child in e.children:
            if isinstance(child, domish.Element):
                ret += xml_pformat(child, indent + 2)
            else:
                ret += '%s%s\n' % (' ' * (indent + 2), child)

        ret += '%s</%s>' % (' ' * indent, e.name)
    else:
        ret += ' />\n'

    return ret

def _hash_capabilities(S):
    r"""
    >>> _hash_capabilities(
    ...     'client/pc//Exodus 0.9.1<http://jabber.org/protocol/caps<http://ja'
    ...     'bber.org/protocol/disco#info<http://jabber.org/protocol/disco#ite'
    ...     'ms<http://jabber.org/protocol/muc<')
    'QgayPKawpkPSDYmwT/WM94uAlu0=\n'
    """
    return base64.encodestring(sha1(S).digest())

def hash_capabilities(identities, features):
    r"""
    >>> hash_capabilities(
    ...    [('client', 'pc', 'Exodus 0.9.1')],
    ...    ["http://jabber.org/protocol/disco#info",
    ...     "http://jabber.org/protocol/disco#items",
    ...     "http://jabber.org/protocol/muc",
    ...     "http://jabber.org/protocol/caps"])
    'QgayPKawpkPSDYmwT/WM94uAlu0=\n'
    """
    S = ""
    for identity in sorted(identities):
        S += "%s/%s//%s<" % identity

    for feature in sorted(features):
        S += "%s<" % feature

    return _hash_capabilities(S)

def elem_find_children(elem, uri, name):
    # TODO: we should probably use this function in more place
    for child in elem.elements():
        if (child.name, child.uri) == (name, uri):
            yield child

def strip_nonspacing_marks(s):
    return ''.join([c for c in unicodedata.normalize('NFD', s)
        if unicodedata.category(c) != 'Mn'])

def alias_match(x, y):
   r"""
   >>> alias_match(u'oo', u'foo')
   True
   >>> alias_match(u'aa', u'foo')
   False

   U+00D3 is LATIN CAPITAL LETTER O WITH ACUTE

   >>> alias_match(u'oo', u'fo\xd3')
   True
   >>> alias_match(u'o\xd3', u'fo\xd3')
   True
   >>> alias_match(u'o\xd3', u'foo')
   True
   """
   return (strip_nonspacing_marks(unicode(x)).lower() in
           strip_nonspacing_marks(unicode(y)).lower())

